import logging
import shutil
import tempfile
from .base import TargetBase
from git import Repo
from git.exc import GitCommandError

LOG = logging.getLogger(__name__)
DEFAULT_MESSAGE = "updating repo programatically (via surgen)"


class GitTarget(TargetBase):
    def __init__(self, target, message=None, commit=True, branch=None):
        super(GitTarget, self).__init__(target)
        self._message = message
        self._workspace = None
        self._commit = commit
        self._branch = branch

    @property
    def workspace(self):
        return self._workspace

    def prepare(self):
        self._workspace = tempfile.mkdtemp()
        self.log("cloning {0}...".format(self.workspace))
        self._repo = Repo.clone_from(self._target, self.workspace)

    def commit(self, summary):
        """
        summary: a summary of the action performed on the target.
        """
        if not self._commit:
            self.log("committing disabled for {0}".format(self))
            return

        message = "{0}\n\n{1}".format(self._message or DEFAULT_MESSAGE, str(summary))

        try:
            self._commit_and_push(self._repo, self._branch, self._message)
        except Exception as e:
            LOG.exception("")
            print(e)
            self.log("unable to push to {0}".format(self._target))

    def _commit_and_push(self, repo, branch, message):
        if repo.is_dirty():
            self.log("committing changes to {0}".format(self._target))
            repo.git.add(A=True)
            repo.git.commit(all=True, message=message)
        self._push(repo, branch)

    def _push(self, repo, branch):
        if branch:
            # this typically means the remote doesn't have the branch to rebase against
            try:
                repo.git.pull(self._repo.remotes[0].name, branch, rebase=True)
            except GitCommandError:
                pass
            repo.git.push(
                self._repo.remotes[0].name,
                "{0}:{1}".format(self._repo.active_branch.name, branch),
            )
        else:
            repo.git.pull(rebase=True)
            repo.git.push()

    def cleanup(self):
        shutil.rmtree(self._workspace)



import os
import py


def find(root_dir, name):
    """
    recurse through the desired directory, and return a py.path.local
    object that matches that name.
    """
    for root, _, file_names in os.walk(root_dir):
        if name in file_names:
            return py.path.local(os.path.join(root, name))



from surgen.surgen import _procedures_from_dir
from .utils import PROCEDURE_DIR


def test_procedures_in_lexical_order():
    """ procedures should return back in lexical order. """
    procedures = _procedures_from_dir(PROCEDURE_DIR)
    procedure_file_list = list(procedures.keys())
    assert procedure_file_list == sorted(procedure_file_list)



import os
from collections import OrderedDict
from .procedure import from_file
from .exceptions import ProcedureNotFound
from .results import ResultStatus, Result, Results

import logging

LOG = logging.getLogger(__name__)


class Surgen(object):
    """ the main context to perform operations. """

    def __init__(self, procedures_by_name):
        self._procedures_by_name = procedures_by_name
        # lists data about performed procedures
        self._performed_procedure_data = []

    def operate(self, target, ignore_errors, dry_run):
        """
        perform the operation on the target directory.
        return an exit code.
        """
        LOG.info("Perfoming procedures on {0}".format(target))
        target.prepare()
        try:
            results = Results()
            for name, procedure_cls in self._procedures_by_name.items():
                LOG.info("  {0}:".format(name))
                status = self._run_procedure(
                    name, procedure_cls, target.workspace, dry_run
                )
                results.add(Result(procedure=name, status=status))
                if not ignore_errors and status == ResultStatus["FAIL"]:
                    LOG.error("Surgen ending early")
                    self._print_results(results)
                    return 1
            if not dry_run:
                target.commit(results)
            self._print_results(results)
            self._performed_procedure_data.append((target, results))
            return 1 if ResultStatus["FAIL"] in results else 0
        finally:
            target.cleanup()

    def _print_results(self, results):
        LOG.notice(
            "Complete! {0} procedure(s) performed.".format(
                len(self._procedures_by_name)
            )
        )
        LOG.notice(
            "{PASS} success, {FAIL} failed, {SKIP} skipped".format(
                **results.count_by_status
            )
        )

    def _run_procedure(self, name, procedure_cls, target_dir, dry_run):
        LOG.info("executing...".format(name))
        procedure = procedure_cls(name, target_dir)
        try:
            should_not_run_reason = procedure.should_not_run()
        except Exception as e:
            LOG.exception("procedure raised an exception during should_not_run! {0}".format(e))
            return ResultStatus["FAIL"]
        if should_not_run_reason:
            LOG.info(
                "skipping, should_not_run returned: {0}".format(
                    should_not_run_reason
                )
            )
            return ResultStatus["SKIP"]
        if not dry_run:
            try:
                result = procedure.operate()
            except Exception as e:
                LOG.exception("procedure raised an exception! {0}".format(e))
                return ResultStatus["FAIL"]
            LOG.info("complete!")
        return ResultStatus["PASS"]

    def print_total_summary(self):
        """ print the total summary. """
        for target, results in self._performed_procedure_data:
            level = logging.SUCCESS
            if results.count_by_status["FAIL"] > 0:
                level = logging.ERROR
            if results.count_by_status["PASS"] == 0:
                level = logging.NOTICE
            LOG.log(level, "{0}: {PASS} | {FAIL} | {SKIP}".format(
                target, **results.count_by_status
            ))


def surgen_from_directory(procedure_dir):
    """ load a surgen object from a directory """
    return Surgen(_procedures_from_dir(procedure_dir))


def _procedures_from_dir(procedure_dir):
    procedures_by_name = OrderedDict()
    for d in sorted(os.listdir(procedure_dir)):
        if os.path.isdir(d):
            LOG.info("skipping {0}, it is a directory.".format(d))
            continue
        if not d.endswith(".py"):
            LOG.info("skipping {0}, it does not end with extension .py".format(d))
            continue
        name = d[: -len(".py")]
        full_path = os.path.join(procedure_dir, d)
        try:
            procedures_by_name[name] = from_file(full_path)
        except ProcedureNotFound as pnf:
            LOG.warn("skipping {0}: {1}".format(d, pnf))
            continue
    return procedures_by_name



import difflib
import logging
import os
import shutil
import tempfile
from pprint import pprint
from surgen.procedure import from_file

DIFFER = difflib.Differ()
LOG = logging.getLogger(__name__)


def compare_procedure(procedure_file, inp, output):
    """
    test a procedure against a filesystem, by taking in:

    :param surgen.Procedure procedure: the procedure to test
    :param str inp: the directory containing the input to operate on
    :param str output: the directory containing the output to validate against
    """
    tmp_dir = tempfile.mkdtemp()
    shutil.rmtree(tmp_dir)
    shutil.copytree(inp, tmp_dir)
    procedure_cls = from_file(procedure_file)
    p = procedure_cls("dummy", tmp_dir)
    assert not p.should_not_run(), "procedure indicated this should not run."
    p.operate()
    _assert_identical_directories(output, tmp_dir)


def _assert_identical_directories(l, r):
    for root, _, file_names in os.walk(l):
        for name in file_names:
            relative_path = os.path.join(root, os.curdir + os.sep + name)[len(l) :]
            # remove absolute at top, if it exists
            relative_path = relative_path.lstrip(os.sep)
            l_path = os.path.join(l, relative_path)
            r_path = os.path.join(r, relative_path)
            with open(l_path) as l_fh:
                with open(r_path) as r_fh:
                    result = list(
                        difflib.ndiff(
                            l_fh.read().splitlines(), r_fh.read().splitlines()
                        )
                    )
                    result = [r for r in result if not r.startswith(" ")]
                    if result:
                        pprint(result)
                    assert not result, "contents of {} != {}.".format(l_path, r_path)



"""surgen, a codebase upgrade tool

Usage:
  surgen [-d -i -v --driver=<driver> --log=<level> <procedure_dir> <target>]

Options:
  <procedure_dir>         the directory containing surgen procedure files [default: ./procedures/]
  <target>                the target to operate on. [default: .]
  --driver=<driver>       the target driver to use. [default: local]
  --log=<level>           log at the specified log level. levels follow those in the verboselogs python library.
  -d, --dry-run           skips execution if provided
  -i, --ignore-errors     continues if a procedure fails
  -h, --help              show this usage guide
"""
import os
import docopt
import logging
import sys
from .surgen import Surgen, surgen_from_directory
from .target import target_from_str, LocalTarget
from .log import setup_logging

DEFAULT_PROCEDURE_DIR = os.path.join(os.curdir, "surgen_procedures")
DEFAULT_TARGET_DIR = os.curdir
DEFAULT_DRIVER = "local"


def main(argv=sys.argv[1:]):
    options = docopt.docopt(__doc__, argv=argv, options_first=True)
    level = options.get("--log") or logging.INFO
    setup_logging(level)
    s = surgen_from_directory(options.get("<procedure_dir>") or DEFAULT_PROCEDURE_DIR)
    d = options.get("--driver") or DEFAULT_DRIVER
    t = (
        target_from_str(d, options.get("<target>"))
        if options.get("<target>")
        else LocalTarget(DEFAULT_TARGET_DIR)
    )
    return s.operate(t, options["--ignore-errors"], options["--dry-run"])



from surgen.results import ResultStatus


def test_result_formatting():
    assert str(ResultStatus["PASS"]) == "PASS"
    assert str(ResultStatus["FAIL"]) == "FAIL"
    assert str(ResultStatus["SKIP"]) == "SKIP"



import attr

ResultStatus = {"PASS": "PASS", "FAIL": "FAIL", "SKIP": "SKIP"}


@attr.s(frozen=True)
class Result(object):
    procedure = attr.ib()
    status = attr.ib()

    def __str__(self):
        return "procedure: {0}, status: {1}".format(self.procedure, self.status)


class Results(list):
    """ stores a list of results. """

    def __init__(self):
        super(Results, self).__init__()
        self._by_status = {}
        for r in ResultStatus.values():
            self._by_status[r] = 0

    def __str__(self):
        return "\n".join(list(str(e) for e in self))

    def add(self, result):
        super(Results, self).append(result)
        self._by_status[result.status] += 1

    @property
    def count_by_status(self):
        return self._by_status



import coloredlogs
import logging

def setup_logging(level):
    level = logging.getLevelName(level)
    root_logger = logging.getLogger()
    root_logger.setLevel(level)
    handler = logging.StreamHandler()
    handler.setLevel(level)
    formatter = coloredlogs.ColoredFormatter(
        "%(asctime)s %(levelname)s %(message)s"
    )
    handler.setFormatter(formatter)
    root_logger.addHandler(handler)



from .git_target import GitTarget
from .local_target import LocalTarget
from ..exceptions import TargetDriverNotFound


def target_from_str(driver, target):
    if driver == "local":
        return LocalTarget(target)
    elif driver == "git":
        return GitTarget(target)

    raise TargetDriverNotFound(driver)



import logging
LOG = logging.getLogger(__name__)

class TargetBase(object):
    TARGET_TYPE = "base"

    def __init__(self, target):
        self._target = target

    def log(self, msg):
        LOG.info(msg)

    @property
    def workspace(self):
        raise NotImplementedError()

    def prepare(self):
        pass

    def commit(self, summary):
        pass

    def cleanup(self):
        pass

    def __str__(self):
        return "{0}: ({1})".format(self.__class__.__name__, self._target)
